from enum import Enum
from pydantic import BaseModel, Field
from typing import Any, List, Optional

from pyrasgo.schemas.attributes import Attribute
from pyrasgo.schemas.data_source import DataSource
from pyrasgo.schemas.granularity import Granularity

class DataType(Enum):
    INT = 'int'
    INTEGER = 'integer'
    FLOAT = 'float'
    FLOAT64 = 'float64'
    DECIMAL = 'decimal'
    NUMERIC = 'numeric'
    NUMBER = 'number'
    REAL = 'real'
    DOUBLE = 'double'
    STRING = 'string'
    TEXT = 'text'
    VARCHAR = 'varchar'
    CHAR = 'char'
    DATE = 'date'
    DATETIME = 'datetime'
    TIMESTAMP = 'timestamp'
    BINARY = 'binary'
    BOOLEAN = 'boolean'
    BOOL = 'bool'

    @classmethod
    def __missing__(cls, key):
        formatted = key.replace(" ", "").lower()
        try:
            return cls._value2member_map_[formatted]
        except KeyError:
            raise ValueError("%r is not a valid %s" % (key, cls.__name__))

# Schemas to be used for general pyrasgo
class Dimension(BaseModel):
    columnName: str
    dataType: DataType
    granularity: Optional[str]

class Feature(BaseModel):
    columnName: str
    dataType: DataType
    description: Optional[str]
    displayName: Optional[str]
    status: Optional[str]
    tags: Optional[List[str]]
    attributes: Optional[List[Attribute]]
    class Config:
        allow_population_by_field_name = True

class FeatureSet(BaseModel):
    id: Optional[int]
    name: Optional[str]
    sourceTable: str
    dimensions: Optional[List[Dimension]]
    features: Optional[List[Feature]]
    dataSource: Optional[DataSource]
    granularities: Optional[List[Granularity]]


# Schema to be used for YML files
# TODO: sync with new schema when customers' legacy files are converted
class DimensionYML(BaseModel):
    columnName: str
    dataType: str
    granularity: str
    class Config:
        allow_population_by_field_name = True

class FeatureYML(BaseModel):
    columnName: str
    dataType: str
    description: Optional[str]
    displayName: Optional[str]
    tags: Optional[List[str]]
    class Config:
        allow_population_by_field_name = True

class DataSourceYML(BaseModel):
    name: Optional[str]

class FeatureSetYML(BaseModel):
    name: Optional[str] = ''
    sourceTable: str
    dimensions: Optional[List[DimensionYML]]
    features: Optional[List[FeatureYML]]
    dataSource: Optional[DataSourceYML]
    script: Optional[str] = ''
    class Config:
        allow_population_by_field_name = True


import json
import logging
import requests
import functools

from pyrasgo.version import __version__ as pyrasgo_version
from pyrasgo.session import Environment

PRODUCTION_HEAP_KEY = '540300130'
STAGING_HEAP_KEY = '3353132567'

HEAP_URL = "https://heapanalytics.com/api"
# HEAP_PROPS_URL = f"{HEAP_URL}/add_user_properties"


def track_usage(func):
    @functools.wraps(func)
    def decorated(self, *args, **kwargs):
        try:
            self._api_key
        except AttributeError:
            logging.debug(f"Cannot track functions called from {self.__class__.__name__} class.")
            return func(self, *args, **kwargs)

        if self._environment == Environment.LOCAL:
            logging.info(f"Called {func.__name__} with parameters: {kwargs}")
        else:
            try:
                track_call(
                    app_id=PRODUCTION_HEAP_KEY if self._environment == Environment.PRODUCTION else STAGING_HEAP_KEY,
                    identity=self._profile.get('id', 0),
                    event=func.__name__,
                    properties={"hostname": self._environment.value,
                                "source": "pyrasgo",
                                "class": self.__class__.__name__,
                                "version": pyrasgo_version,
                                "userId": self._profile.get('id', 0),
                                "username": self._profile.get('username', 'Unknown'),
                                "orgId": self._profile.get('organizationId', 0),
                                "input": args,
                                **kwargs})
            except Exception:
                logging.debug(f"Called {func.__name__} with parameters: {kwargs}")
        return func(self, *args, **kwargs)

    return decorated


def identify(app_id: str,
             identity: int):
    """
    Send an "identify" event to the Heap Analytics API server
    
    :param identity: unique id used to identify the user
    """
    data = {"app_id": app_id,
            "identity": identity}

    response = requests.post(url=f"{HEAP_URL}/identify",
                             data=json.dumps(data),
                             headers={"Content-Type": "application/json"})
    response.raise_for_status()
    return response


def add_user_properties(app_id: str,
                        identity: int,
                        properties: dict = None):
    """
    Send a "add_user_properties" event to the Heap Analytics API server
    
    :pram: identity: unique id used to identify the user
    :param properties: optional, additional properties to associate with the user
    """
    data = {"app_id": app_id,
            "identity": identity}
    
    if properties is not None:
        data["properties"] = properties

    response = requests.post(url=f"{HEAP_URL}/add_user_properties",
                             data=json.dumps(data),
                             headers={"Content-Type": "application/json"})
    response.raise_for_status()
    return response


def track_call(app_id: str,
               identity: int,
               event: str,
               properties: dict = None):
    """
    Send a "track" event to the Heap Analytics API server.

    :param identity: unique id used to identify the user
    :param event: event name
    :param properties: optional, additional event properties
    """
    data = {"app_id": app_id,
            "identity": identity,
            "event": event}

    if properties is not None:
        data["properties"] = properties

    response = requests.post(url=f"{HEAP_URL}/track",
                             data=json.dumps(data),
                             headers={"Content-Type": "application/json"})
    response.raise_for_status()
    return response



from typing import Union, List

import time
from tqdm import tqdm

from pyrasgo.connection import Connection
from pyrasgo.enums import ModelType
from pyrasgo.feature import Feature, FeatureList
from pyrasgo.monitoring import track_usage
from pyrasgo.namespace import Namespace
from pyrasgo.schemas import attributes as att


class Model(Connection):
    """
    Stores all the features for a model
    """

    def __init__(self, api_object, **kwargs):
        self._namespace = Namespace(**api_object)
        super().__init__(**kwargs)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Model(id={self.id}, name={self.name}, authorId={self.authorId})"

    def __getattr__(self, item):
        try:
            return self._namespace.__getattribute__(item)
        except KeyError:
            self.get()
        try:
            return self._namespace.__getattribute__(item)
        except KeyError:
            raise AttributeError(f"No attribute named {item}")

    @property
    @track_usage
    def type(self):
        """
        Returns the model type of the model
        """
        return ModelType(self._namespace.type)


    @property
    @track_usage
    def features(self):
        """
        Returns the features within the model
        """
        try:
            return FeatureList([feature.dict() for feature in self._namespace.features])
        except AttributeError:
            self.get()
        return FeatureList([feature.dict() for feature in self._namespace.features])

    @features.setter
    def features(self, features: Union[List[Feature], FeatureList, Feature]):
        # Update the namespace
        self._namespace.__setattr__('features',
                                    [Namespace(**feature.dict()) for feature in self.features + features])

    def get(self):
        """
        Updates the Model object's attributes from the API
        """
        self._namespace = Namespace(**self._get(f"/models/{self.id}", api_version=1).json())

    def get_author(self):
        """
        Returns the author object of the model (includes credentials if user is author)
        """
        if self._profile['id'] == self.authorId:
            author = self._profile
        else:
            author = self._get(f"/users/{self.authorId}", api_version=1).json()
        return f'Author({author["id"]}, {author["firstName"]} {author["lastName"]})'


    @track_usage
    def share(self, share: bool=True) -> bool:
        """
        Share the model with your organization
        :param share: True to make your model public. False to make your model private
        """
        response = self._patch(f"/models/{self.id}/details", api_version=1, _json={"isShared": share}).json()
        return response['isShared']



    @track_usage
    def add_feature(self, feature: Feature) -> None:
        """
        Adds a single feature to the model
        """
        self.features = feature
        self._patch(f"/models/{self.id}/features", api_version=1,
                    _json={"featureIds": [str(feature.id)]})

    @track_usage
    def add_features(self, features: Union[List[Feature], FeatureList]) -> None:
        """
        Adds a FeatureList or a list of Features to the model.
        """
        self.features = features
        self._patch(f"/models/{self.id}/features", api_version=1,
                    _json={"featureIds": [str(feature.id) for feature in features]})

    @track_usage
    def is_data_ready(self) -> bool:
        """
        Performs check against API for training data readiness, if true, a dataframe can be pulled down.
        :return:
        """
        r = self._get("/trainings/latest", api_version=0)
        if any([model['state'] == 'done' for model in r.json() if model['model']['id'] == self.id]):
            return True
        if any([model['state'] == 'error' for model in r.json() if model['model']['id'] == self.id]):
            raise SystemError("There's been an issue with the training data generation")
        return False

    @track_usage
    def generate_training_data(self) -> None:
        """
        Triggers the generation of the model's training data.
        """
        self._post(f"/trainings/", _json={"modelId": int(self.id),
                                          "userId": self._profile['id']},
                   api_version=0)
        for _ in tqdm(range(0, 10), leave=False):
            if self.is_data_ready():
                break
            time.sleep(1)

    @track_usage
    def add_attributes(self, attributes: List[dict]):
        if not isinstance(attributes, list):
            raise ValueError('attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]')
        attr = []
        for kv in attributes:
            if not isinstance(kv, dict):
                raise ValueError('attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]')
            for k, v in kv.items():
                attr.append(att.Attribute(key=k, value=v))
        attr_in = att.CollectionAttributeBulkCreate(collectionId = self.id, attributes=attr)
        return self._put(f"/models/{self.id}/attributes", attr_in.dict(exclude_unset=True), api_version=1).json()

    @track_usage
    def get_attributes(self):
        try:
            response = self._get(f"/models/{self.id}/attributes", api_version=1).json()
        except:
            return 'Cannot find attributes for this collection'
        dict_out = {}
        for kv in response:
            dict_out[kv['key']] = kv['value']
        return att.CollectionAttributes(collectionId=self.id, attributes=dict_out)

    def snowflake_table_metadata(self, creds):
        metadata = {
            "database": creds.get("database", "rasgo"),
            "schema": creds.get("schema", "public"),
            "table": self._snowflake_table_name(),
        }
        return metadata

    def _snowflake_table_name(self):
        table_name = self.dataTableName
        if table_name is None:
            raise ValueError("No table found for model '{}'".format(self.id))
        return table_name



from pydantic import BaseModel, Field
from typing import List

class Attribute(BaseModel):
    key: str
    value: str

class FeatureAttributes(BaseModel):
    featureId: int
    attributes: dict

class FeatureAttributesLog(BaseModel):
    featureId: int
    attributes: List[dict]

class FeatureAttributeBulkCreate(BaseModel):
    featureId: int
    attributes: List[Attribute]

class CollectionAttributes(BaseModel):
    modelId: int = Field(alias='collectionId')
    attributes: dict

class CollectionAttributesLog(BaseModel):
    modelId: int = Field(alias='collectionId')
    attributes: List[dict]

class CollectionAttributeBulkCreate(BaseModel):
    modelId: int = Field(alias='collectionId')
    attributes: List[Attribute]


from .attributes import (Attribute, 
                         FeatureAttributes, FeatureAttributesLog, FeatureAttributeBulkCreate,
                         CollectionAttributes, CollectionAttributesLog, CollectionAttributeBulkCreate)
from .column import Column, ColumnCreate, ColumnUpdate
from .data_source import DataSource, DataSourceCreate, DataSourceUpdate
from .dimensionality import Dimensionality, DimensionalityCreate
from .feature import FeatureCreate, FeatureUpdate, FeatureStats
from .feature_set import v0, v1
from .organization import Organization



from typing import List, Optional
from pydantic import BaseModel

from pyrasgo.schemas.attributes import Attribute
from pyrasgo.schemas.granularity import Granularity
from pyrasgo.schemas.feature_set import v0


class FeatureBase(BaseModel):
    id: int


class FeatureCreate(BaseModel):
    name: str
    code: Optional[str]
    description: Optional[str]
    columnId: int
    featureSetId: Optional[int]
    organizationId: int
    orchestrationStatus: Optional[str]
    tags: Optional[List[str]]
    gitRepo: Optional[str]

class FeatureUpdate(BaseModel):
    id: int
    name: Optional[str]
    code: Optional[str]
    description: Optional[str]
    gitRepo: Optional[str]
    #Can't include these until we create a v1 patch endpoint for features
    #columnId: Optional[int]
    #featureSetId: Optional[int]
    #organizationId: Optional[int]
    orchestrationStatus: Optional[str]
    #tags: Optional[List[str]]

class Feature(FeatureBase):
    id: str  # TODO: Returning feature ids as strings, ensure consistency.
    name: str
    description: Optional[str]
    code: Optional[str]
    columnId: Optional[int]
    dataType: Optional[str]
    featureSet: Optional[v0.FeatureSet]
    granularities: Optional[List[Granularity]]
    orchestrationStatus: Optional[str]
    tags: Optional[List[str]]
    attributes: Optional[List[Attribute]]
    gitRepo: Optional[str]
    organizationId: Optional[int]

class FeatureStats(BaseModel):
    recCt: Optional[int]
    distinctCt: Optional[int]
    nullRecCt: Optional[int]
    zeroValRecCt: Optional[int]
    meanVal: Optional[str]
    medianVal: Optional[str]
    maxVal: Optional[str]
    minVal: Optional[str]
    sumVal: Optional[str]
    stdDevVal: Optional[str]
    varianceVal: Optional[str]
    rangeVal: Optional[str]
    skewVal: Optional[str]
    kurtosisVal: Optional[str]
    q1Val: Optional[str]
    q3Val: Optional[str]
    IQRVal: Optional[str]
    pct5Val: Optional[str]
    pct95Val: Optional[str]
    outlierCt: Optional[str]
    lowOutlier: Optional[str]
    highOutlier: Optional[str]